@@ -6,17 +6,18 @@
 #include <stdlib.h>
 #include <sys/types.h>
 #include <dirent.h>
+#include <time.h>
 
 #include "libacpi.h"
 
 extern char *state[];
-extern global_t *globals;
+/* extern global_t *globals; */
 
 /* local proto */
 int acpi_get_design_cap(int batt);
 
 /* initialise the batteries */
-int init_batteries(void)
+int init_batteries(global_t *globals)
 {
     DIR *battdir;
     struct dirent *batt;
@@ -83,16 +84,16 @@ int init_batteries(void)
 }
 
 /* a stub that just calls the current function */
-int reinit_batteries(void)
+int reinit_batteries(global_t *globals)
 {
     pdebug("reinitialising batteries\n");
-    return init_batteries();
+    return init_batteries(globals);
 }
 
 /* the actual name of the subdirectory under ac_adapter may
  * be anything, so we need to read the directory and use the
  * name we find there. */
-int init_ac_adapters(void)
+int init_ac_adapters(global_t *globals)
 {
     DIR *acdir;
     struct dirent *adapter;
@@ -125,14 +126,14 @@ int init_ac_adapters(void)
 }
 
 /* stub that does nothing but call the normal init function */
-int reinit_ac_adapters(void)
+int reinit_ac_adapters(global_t *globals)
 {
     pdebug("reinitialising ac adapters\n");
-    return init_ac_adapters();
+    return init_ac_adapters(globals);
 }
 
 /* see if we have ACPI support and check version */
-int power_init(void)
+int power_init(global_t *globals)
 {
     FILE *acpi;
     char buf[4096];
@@ -156,14 +157,14 @@ int power_init(void)
     /* yep, all good */
     fclose(acpi);
 
-    if (!(retval = init_batteries()))
-	retval = init_ac_adapters();
+    if (!(retval = init_batteries(globals)))
+	retval = init_ac_adapters(globals);
 
     return retval;
 }
 
 /* reinitialise everything, to deal with changing batteries or ac adapters */
-int power_reinit(void)
+int power_reinit(global_t *globals)
 {
     FILE *acpi;
     int retval;
@@ -173,8 +174,8 @@ int power_reinit(void)
 	return 1;
     }
     
-    if (!(retval = reinit_batteries()))
-	retval = reinit_ac_adapters();
+    if (!(retval = reinit_batteries(globals)))
+	retval = reinit_ac_adapters(globals);
 
     return retval;
 }
@@ -195,7 +196,14 @@ char *get_value(char *string)
     return retval;
 }
 
-power_state_t get_power_status(void)
+int check_error(char *buf)
+{
+    if(strstr(buf, "ERROR") != NULL)
+	return 1;
+    return 0;
+}
+
+power_state_t get_power_status(global_t *globals)
 {
     FILE *file;
     char buf[1024];
@@ -237,6 +245,13 @@ int get_battery_info(int batt_no)
     buflen = fread(buf, sizeof(buf), 1, file);
     fclose(file);
 
+    /* check to see if there were any errors reported in the file */
+    if(check_error(buf)) {
+	pinfo("Error reported in file %s - discarding data\n",
+	      info->info_file);
+	return 0;
+    }
+
     /* check to see if battery is present */
     entry = strstr(buf, "present:");
     val = get_value(entry);
@@ -282,9 +297,16 @@ int get_battery_info(int batt_no)
     }
     
     /* grab the file contents */
+    memset(buf, 0, sizeof(buf));
     buflen = fread(buf, sizeof(buf), 1, file);
     fclose(file);
 
+    /* check to see if there were any errors reported in the file */
+    if(check_error(buf)) {
+	pinfo("Error reported in file %s - discarding data\n",
+	      info->state_file);
+	return 0;
+    }
     /* check to see if battery is present */
     entry = strstr(buf, "present:");
     val = get_value(entry);
@@ -386,13 +408,48 @@ static int calc_remaining_percentage(int batt)
     return retval;
 }
 
+/* check to see if we've been getting bad data from the batteries - if
+ * we get more than some limit we switch to using the remaining capacity
+ * for the calculations. */
+static enum rtime_mode check_rt_mode(global_t *globals)
+{
+    int i;
+    int bad_limit = 5;
+    battery_t *binfo;
+
+    /* if we were told what to do, we should keep doing it */
+    if(globals->rt_forced)
+	return globals->rt_mode;
+
+    for(i = 0; i < MAXBATT; i++) {
+	binfo = &batteries[i];
+	if(binfo->present && globals->adapter.power == BATT) {
+	    if(binfo->present_rate <= 0) {
+		pdebug("Bad report from %s\n", binfo->name);
+		binfo->bad_count++;
+	    }
+	}
+    }
+    for(i = 0; i < MAXBATT; i++) {
+	binfo = &batteries[i];
+	if(binfo->bad_count > bad_limit) {
+	    if(globals->rt_mode != RT_CAP)
+		pinfo("More than %d bad reports from %s; "
+		      "Switching to remaining capacity mode\n",
+		      bad_limit, binfo->name);
+	    return RT_CAP;
+	}
+    }
+    return RT_RATE;
+}
+
 /* calculate remaining time until the battery is charged.
  * when charging, the battery state file reports the 
  * current being used to charge the battery. We can use 
  * this and the remaining capacity to work out how long
  * until it reaches the last full capacity of the battery.
  * XXX: make sure this is actually portable . . . */
-static int calc_charge_time(int batt)
+static int calc_charge_time_rate(int batt)
 {
     float rcap, lfcap;
     battery_t *binfo;
@@ -416,7 +473,78 @@ static int calc_charge_time(int batt)
     return charge_time;
 }
 
-void acquire_batt_info(int batt)
+/* we need to calculate the present rate the same way we do in rt_cap
+ * mode, and then use that to estimate charge time. This will 
+ * necessarily be even less accurate than it is for remaining time, but
+ * it's just as neessary . . . */
+static int calc_charge_time_cap(int batt)
+{
+    static float cap_samples[CAP_SAMPLES];
+    static int time_samples[CAP_SAMPLES];
+    static int sample_count = 0;
+    static int current = 0;
+    static int old = 1;
+    int rtime;
+    int tdiff;
+    float cdiff;
+    float current_rate;
+    battery_t *binfo = &batteries[batt];
+    
+    cap_samples[current] = (float) binfo->remaining_cap;
+    time_samples[current] = time(NULL);
+
+    if (sample_count == 0) {
+	/* we can't do much if we don't have any data . . . */
+	current_rate = 0;
+    } else if (sample_count < CAP_SAMPLES) {
+	/* if we have less than SAMPLES samples so far, we use the first
+	 * sample and the current one */
+	cdiff = cap_samples[current] - cap_samples[0];
+	tdiff = time_samples[current] - time_samples[0];
+	current_rate = cdiff/tdiff;
+    } else {
+	/* if we have more than SAMPLES samples, we use the oldest
+	 * current one, which at this point is current + 1. This will
+	 * wrap the same way that current will wrap, but one cycle
+	 * ahead */
+	cdiff = cap_samples[current] - cap_samples[old];
+	tdiff = time_samples[current] - time_samples[old];
+	current_rate = cdiff/(float)tdiff;
+    }
+    if (current_rate == 0) 
+	rtime = 0;
+    else {
+	float cap_left = (float)(binfo->last_full_cap - binfo->remaining_cap);
+	rtime = (int)(cap_left/(current_rate * 60.0));
+    }
+    sample_count++, current++, old++;
+    if (current >= CAP_SAMPLES)
+	current = 0;
+    if (old >= CAP_SAMPLES)
+	old = 0;
+
+    pdebug("cap charge time rem: %d\n", rtime);
+    return rtime;
+}
+
+static int calc_charge_time(global_t *globals, int batt)
+{
+    int ctime = 0;
+
+    globals->rt_mode = check_rt_mode(globals);
+
+    switch(globals->rt_mode) {
+    case RT_RATE:
+	ctime = calc_charge_time_rate(batt);
+	break;
+    case RT_CAP:
+	ctime = calc_charge_time_cap(batt);
+	break;
+    }
+    return ctime;
+}
+
+void acquire_batt_info(global_t *globals, int batt)
 {
     battery_t *binfo;
     adapter_t *ap = &globals->adapter;
@@ -451,7 +579,7 @@ void acquire_batt_info(int batt)
 
     /* we need to /know/ that we've got a valid state for the 
      * globals->power value . . . .*/
-    ap->power = get_power_status();
+    ap->power = get_power_status(globals);
 
     if ((ap->power != AC) && (binfo->charge_state == DISCHARGE)) {
 	/* we're not on power, and not charging. So we might as well 
@@ -463,32 +591,51 @@ void acquire_batt_info(int batt)
 	}
     }
     
-    binfo->charge_time = calc_charge_time(batt);
+    binfo->charge_time = calc_charge_time(globals, batt);
 
     /* and finally, we tell anyone who wants to use this information
      * that it's now valid . . .*/
     binfo->valid = 1;
 }
 	
-void acquire_all_batt_info(void)
+void acquire_all_batt_info(global_t *globals)
 {
     int i;
     
     for(i = 0; i < globals->battery_count; i++)
-	acquire_batt_info(i);
+	acquire_batt_info(globals, i);
 }
 
-void acquire_global_info(void)
+/*
+ * One of the feature requests I've had is for some way to deal with
+ * batteries that are too dumb or too b0rken to report a present rate
+ * value. The way to do this, obviously, is to record the time that
+ * samples were taken and use that information to calculate the rate
+ * at which the battery is draining/charging. This still won't help
+ * systems where the battery doesn't even report the remaining
+ * capacity, but without the present rate or the remaining capacity, I
+ * don't think there's /anything/ we can do to work around it.
+ *
+ * So, what we need to do is provide a way to use a different method
+ * to calculate the time remaining. What seems most sensible is to
+ * split out the code to calculate it into a seperate function, and
+ * then provide multiple implementations . . . 
+ */
+
+/*
+ * the default implementation - if present rate and remaining capacity
+ * are both reported correctly, we use them.
+ */
+int calc_time_remaining_rate(global_t *globals)
 {
     int i;
     int rtime;
     float rcap = 0;
     float rate = 0;
     battery_t *binfo;
-    adapter_t *ap = &globals->adapter;
     static float rate_samples[SAMPLES];
-    static int j = 0;
     static int sample_count = 0;
+    static int j = 0;
     static int n = 0;
 
     /* calculate the time remaining, using the battery's remaining 
@@ -508,7 +655,8 @@ void acquire_global_info(void)
     }
     rate_samples[j] = rate;
     j++, sample_count++;
-    j = j % SAMPLES;
+    if (j >= SAMPLES)
+	j = 0;
     
     /* for the first SAMPLES number of calls we calculate the
      * average based on sample_count, then we use SAMPLES to
@@ -540,17 +688,105 @@ void acquire_global_info(void)
     if(rtime <= 0) 
 	rtime = 0;
  out:
-    pdebug("time rem: %d\n", rtime);
-    globals->rtime = rtime;
+    pdebug("discharge time rem: %d\n", rtime);
+    return rtime;
+}
+
+/*
+ * the alternative implementation - record the time at which each
+ * sample was taken, and then use the difference between the latest
+ * sample and the one SAMPLES ago to calculate the difference over
+ * that time, and from there the rate of change of capacity.
+ *
+ * XXX: this code sucks, but largely because batteries aren't exactly
+ * precision instruments - mine only report with about 70mAH
+ * resolution, so they don't report any changes until the difference
+ * is 70mAH. This means that calculating the current rate from the
+ * remaining capacity is very choppy . . . 
+ *
+ * To fix this, we should calculate an average over some number of
+ * samples at the old end of the set - this would smooth out the
+ * transitions. 
+ */
+int calc_time_remaining_cap(global_t *globals)
+{
+    static float cap_samples[CAP_SAMPLES];
+    static int time_samples[CAP_SAMPLES];
+    static int sample_count = 0;
+    static int current = 0;
+    static int old = 1;
+    battery_t *binfo;
+    int i;
+    int rtime;
+    int tdiff;
+    float cdiff;
+    float cap = 0;
+    float current_rate;
+
+    for (i = 0; i < globals->battery_count; i++) {
+	binfo = &batteries[i];
+	if (binfo->present && binfo->valid)
+	    cap += binfo->remaining_cap;
+    }
+    cap_samples[current] = cap;
+    time_samples[current] = time(NULL);
+
+    if (sample_count == 0) {
+	/* we can't do much if we don't have any data . . . */
+	current_rate = 0;
+    } else if (sample_count < CAP_SAMPLES) {
+	/* if we have less than SAMPLES samples so far, we use the first
+	 * sample and the current one */
+	cdiff = cap_samples[0] - cap_samples[current];
+	tdiff = time_samples[current] - time_samples[0];
+	current_rate = cdiff/tdiff;
+    } else {
+	/* if we have more than SAMPLES samples, we use the oldest
+	 * current one, which at this point is current + 1. This will
+	 * wrap the same way that current will wrap, but one cycle
+	 * ahead */
+	cdiff = cap_samples[old] - cap_samples[current];
+	tdiff = time_samples[current] - time_samples[old];
+	current_rate = cdiff/tdiff;
+    }
+    if (current_rate == 0) 
+	rtime = 0;
+    else
+	rtime = (int)(cap_samples[current]/(current_rate * 60.0));
+
+    sample_count++, current++, old++;
+    if (current >= CAP_SAMPLES)
+	current = 0;
+    if (old >= CAP_SAMPLES)
+	old = 0;
+
+    pdebug("cap discharge time rem: %d\n", rtime);
+    return rtime;
+}    
+
+void acquire_global_info(global_t *globals)
+{
+    adapter_t *ap = &globals->adapter;
+
+    globals->rt_mode = check_rt_mode(globals);
+
+    switch(globals->rt_mode) {
+    case RT_RATE:
+	globals->rtime = calc_time_remaining_rate(globals);
+	break;
+    case RT_CAP:
+	globals->rtime = calc_time_remaining_cap(globals);
+	break;
+    }
 
     /* get the power status.
      * note that this is actually reported seperately from the
      * battery info, under /proc/acpi/ac_adapter/AC/state */
-    ap->power = get_power_status();
+    ap->power = get_power_status(globals);
 }
 
-void acquire_all_info(void)
+void acquire_all_info(global_t *globals)
 {
-    acquire_all_batt_info();
-    acquire_global_info();
+    acquire_all_batt_info(globals);
+    acquire_global_info(globals);
 }
